# These functions can be used for reprojection and compression of tiff files
from osgeo import gdal, ogr
import os
import shutil
import json

gdal.UseExceptions()
base = "full_test/TIFFs/5m/orig/"


def compression(folder, out_folder):
    """
    Args:
        folder: folder with tiff files to be reprojected
        out_folder: folder to store reprojected files

    Results:
        Reprojects, compresses, and stores each tiff file
    """

    for file in os.listdir(folder):
        if not file.endswith(".tif"):
            continue
        print(f"Compressing {file}...", end="")
        try:
            gdal.Translate(out_folder + file, folder + file, strict=True, resampleAlg="bilinear",
                           creationOptions=["COMPRESS=LZW"])
        except Exception as e:
            print(f"\nError occurred: {e}")
        print(f"done!")


def report_projections(folders=["/home/areed-comeaux/Desktop/Projects/summitsnapping/full_test/TIFFs/5m/orig/EastAleutians/"]):
    """
    Args:
        folders - list of folder(s) with tiff files to be checked

    Results:
        Creates a text file listing the types of projections found for a specific folder; text file is saved to said
        folder.
    """
    # TODO needs to be be tested for output location stuff
    for folder in folders:
        projections = {}
        file_count = 0
        error_count = 0
        print("Working on ", folder)
        """for file in os.listdir(folder):
            if not file.endswith('.tif'):
                continue
            raster = gdal.Open(folder + file)
            proj = raster.GetProjection()
            if proj not in projections.keys():
                projections.update({proj: [file]})
            else:
                projections[proj].append(file)
            file_count += 1"""
        out = folder + folder.replace(base, '')[:-1] + "projs.txt"
        with open(out, 'w') as out:
            for key in projections:
                out.write(f"\nProjection: {key}\nRatio: {len(projections[key])}/{file_count}\nFiles:\n")
                for item in projections[key]:
                    out.write(f"{folder}{item}\n")
            out.write(f"Errors: {error_count}")


def reproject(folders, outFolders, epsg='EPSG:3338'):
    """
    Args:
        folders: folders with tiffs to be reprojected
        outFolders: folders to store the reprojected tiffs
        epsg: EPSG code for destination SRS
    Results:
        makes and saves a reprojected tiff file
    """
    """folders = [base+"chugach/", base+"chugach2/", base+"Mid-Acc/",
               base+"Miscellaneous/", base+"Summer/", base+"ifsarKenai/"]
    with open(base+'reproj.txt') as file:
        files = file.readlines()
    files = [i.replace('\n', '') for i in files]
    outFolders = [folder+"reproj/" for folder in folders]"""

    for folder, outFolder in zip(folders, outFolders):
        if not os.path.exists(outFolder):
            os.mkdir(outFolder)
        for file in os.listdir(folder):
            if file.endswith(".tif") or file.endswith(".laz"):
                print(f"Processing {file}...", end=" ")
                src_path = folder + file
                dst_path = outFolder + file
                try:
                    ds = gdal.Open(src_path)
                    if src_path in files:
                        try:
                            print("Reprojecting...", end=" ")
                            gdal.Warp(dst_path, ds, dstSRS='EPSG:3338')
                        except Exception as e:
                            print(f"\nException ocurred: {e}")
                            with open("TIFF Errors.txt", 'a') as error_file:
                                error_file.write(f"{file} : {e}\n")
                    else:
                        shutil.copy(src_path, dst_path)
                    print("Done!")
                except Exception as e:
                    print(f"\nException occurred: {e}")
                    with open("TIFF Errors.txt", 'w') as error_file:
                        error_file.write(f"{file} : {e}\n")

report_projections()